﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Primitives3D;
using XNA4;

namespace XNA4Primitives3DEffect
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont spriteFont;

        struct SaveStateMode
        {
            BlendState m_Blenstate;
            DepthStencilState m_DepthState;
            RasterizerState m_RasterState;
            SamplerState m_SamplerState;

            GraphicsDevice m_Device;

            public SaveStateMode(GraphicsDevice device)
            {
                m_Device = device;

                m_Blenstate = m_Device.BlendState;
                m_DepthState = m_Device.DepthStencilState;
                m_RasterState = m_Device.RasterizerState;
                m_SamplerState = m_Device.SamplerStates[0];
            }

            public void Restore()
            {
                m_Device.BlendState = m_Blenstate;
                m_Device.DepthStencilState = m_DepthState;
                m_Device.RasterizerState = m_RasterState;
                m_Device.SamplerStates[0] = m_SamplerState;
            }
        }

        /// <summary>
        /// input
        /// </summary>
        static InputHandler inputs;
        /// <summary>
        /// FPS Counter
        /// </summary>
        static FpsCounter fpsCounter;

        // effect
        Effect effect;
        Matrix projection;
        Matrix view;
        Matrix world;

        // Store a list of primitive models, plus which one is currently selected.
        List<GeometricPrimitive> primitives = new List<GeometricPrimitive>();
        int currentPrimitiveIndex = 0;

        // shader#
        int currentTechniqueIndex = 0;

        // rotate & zoom
        float angle = 0.0f;
        float zoom = 1.0f;

        // store a wireframe rasterize state
        RasterizerState wireFrameState;

        // Are we rendering in wireframe mode?
        bool isWireframe = true;
        bool isAnimation = true;

        // ambient, diffuse, specular light color
        Vector4 ambient = Color.DarkSlateGray.ToVector4(); 
        Vector4 diffuse = Color.CornflowerBlue.ToVector4();
        Vector4 specular = Color.White.ToVector4();
        Vector3 lightDirection = new Vector3(30, 30, 30);
        Vector3 eyePosition = new Vector3(0.0f, 0.0f, 4.0f);  // same as LookAt

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here
            /// <summary>
            /// inputs component
            /// </summary>
            inputs = new InputHandler(this);
            Components.Add(inputs);
            /// <summary>
            /// fps component
            /// </summary>
            fpsCounter = new FpsCounter(this);
            Components.Add(fpsCounter);

            InitializeTransform();
            InitializeEffect();

            base.Initialize();
        }

        /// <summary>
        /// Initializes the transforms used for the scene
        /// </summary>
        private void InitializeTransform()
        {
            float tilt = (float)Math.PI / 8.0f;
            // Use the world matrix to tilt the cube along x and y axes.
            world = Matrix.CreateRotationX(tilt) * Matrix.CreateRotationY(tilt);
            view = Matrix.CreateLookAt(eyePosition, Vector3.Zero, Vector3.Up);
            projection = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.PiOver4,  // so this is 45 degrees.
                (float)GraphicsDevice.Viewport.Width /
                (float)GraphicsDevice.Viewport.Height,
                1.0f, 100.0f);
        }

        /// <summary>
        /// Initializes the effect
        /// </summary>
        private void InitializeEffect()
        {
            effect = Content.Load<Effect>("AmbientDiffuseSpecularLight");
            effect.Parameters["World"].SetValue(world);
            effect.Parameters["View"].SetValue(view);
            effect.Parameters["Projection"].SetValue(projection);
            effect.CurrentTechnique = effect.Techniques["Technique1"];
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>("hudfont");

            // TODO: use this.Content to load your game content here
            primitives.Add(new CubePrimitive(GraphicsDevice));
            primitives.Add(new SpherePrimitive(GraphicsDevice));
            primitives.Add(new CylinderPrimitive(GraphicsDevice));
            primitives.Add(new TorusPrimitive(GraphicsDevice));
            primitives.Add(new PyramidPrimitive(GraphicsDevice));

            wireFrameState = new RasterizerState()
            {
                FillMode = FillMode.WireFrame,
                CullMode = CullMode.None,
            };
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // TODO: Add your update logic here
            UpdateInput();

            base.Update(gameTime);
        }

        #region UpdateInput
        protected void UpdateInput()
        {
            if (inputs.PauseGame)
            {
                this.Exit();
            }

            // Toggle animation?
            if (inputs.IsKeyPressed(Keys.A))
            {
                isAnimation = !isAnimation;
            }

            // Toggle wireframe?
            if (inputs.IsKeyPressed(Keys.W))
            {
                isWireframe = !isWireframe;
            }

            // Toggle technique?
            if (inputs.IsKeyPressed(Keys.T))
            {
                currentTechniqueIndex = (currentTechniqueIndex + 1) % 3;
            }

            // Change primitive?
            if (inputs.IsKeyPressed(Keys.Space))
            {
                currentPrimitiveIndex = (currentPrimitiveIndex + 1) % primitives.Count;
            }

            // Change Y angle?
            if (inputs.IsKeyPressed(Keys.X))
            {
                angle += 5.0f;
            }

            // Change Translate?
            if (inputs.IsKeyPressed(Keys.Y))
            {
                zoom -= 0.1f;
            }
        }
        #endregion

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.Black);

            if (isWireframe)
            {
                GraphicsDevice.RasterizerState = wireFrameState;
            }
            else
            {
                GraphicsDevice.RasterizerState = RasterizerState.CullCounterClockwise;
            }

            // world matrix
            if (isAnimation) angle += 0.8f;
            world = Matrix.CreateFromYawPitchRoll(MathHelper.ToRadians(angle), 45.0f, 0.0f) * Matrix.CreateTranslation(0, 0, zoom);

            // Draw the current primitive.
            GeometricPrimitive currentPrimitive = primitives[currentPrimitiveIndex];

            // effect
            effect.Parameters["World"].SetValue(world);
            effect.Parameters["AmbientColor"].SetValue(ambient);
            effect.Parameters["DiffuseColor"].SetValue(diffuse);
            lightDirection.Normalize();
            effect.Parameters["LightDirection"].SetValue(lightDirection);
            effect.Parameters["SpecularColor"].SetValue(specular);
            effect.Parameters["SpecularPower"].SetValue(16);
            effect.Parameters["SpecularIntensity"].SetValue(1);
            effect.Parameters["EyePosition"].SetValue(eyePosition);
            if (currentTechniqueIndex == 0)
                effect.CurrentTechnique = effect.Techniques["Technique1"];
            else if (currentTechniqueIndex == 1)
                effect.CurrentTechnique = effect.Techniques["Technique2"];
            else if (currentTechniqueIndex == 2)
                effect.CurrentTechnique = effect.Techniques["Technique3"];

            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Apply();
                currentPrimitive.Draw(effect);
            }

            http://blogs.msdn.com/b/shawnhar/archive/2010/06/18/spritebatch-and-renderstates-in-xna-game-studio-4-0.aspx
            //If you are mixing 3D rendering with 2D objects using SpriteBatch, you may notice that your 3D graphics no longer draw correctly after you have rendered sprites. 
            //This is because the SpriteBatch changes several device states to values that may not be appropriate for drawing in 3D. 
            SaveStateMode mode = new SaveStateMode(GraphicsDevice);
            // Draw text
            spriteBatch.Begin();
            if (currentTechniqueIndex == 0)
                spriteBatch.DrawString(spriteFont, "Technique1", new Vector2(10, 10), Color.White);
            else if (currentTechniqueIndex == 1)
                spriteBatch.DrawString(spriteFont, "Technique2", new Vector2(10, 10), Color.White);
            else if (currentTechniqueIndex == 2)
                spriteBatch.DrawString(spriteFont, "Technique3", new Vector2(10, 10), Color.White);
            spriteBatch.End();
            mode.Restore();

            base.Draw(gameTime);
        }
    }
}
